#ifndef FILE_INTERSECT
#define FILE_INTERSECT
/*
    This header file is part of CadCam.

    This header file is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This header file is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CadCam.  If not, see <https://www.gnu.org/licenses/>.

    AUTOR     : Grotius
    DATE      : 12.6.2020
    REVISION  : --
    TODO      : --
    USAGE     : To use this header only library/file, include also structs.h to your project.
    PROTOCOL  : --
    NOTES     : --
*/
#include <iostream>
#include <algorithm>
#include <math.h>
#include <structs.h>
void arc_angle_it(std::vector<OBJECT>::iterator it);

void lsl_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    std::vector<POINT>::iterator it2;
    it2=it1->control.end();
    //linestrip 1
    double x1=it->control.back().x;
    double y1=it->control.back().y;
    double x2=it->end.x;
    double y2=it->end.y;
    //line 2
    double x3=it1->start.x;
    double y3=it1->start.y;
    double x4=it1->end.x;
    double y4=it1->end.y;

    int determinant=(y2-y4)*(x3-x2)-(x2-x4)*(y3-y2);
    if(determinant==0){
        //skip, lines are collineair
        std::cout<<"linestrip-line is collineair"<<std::endl;
    } else {
        it->end.x=it1->start.x=((((x1*y2)-(y1*x2))*(x3-x4))-((x1-x2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4))); //wikipedia line-line intersection
        it->end.y=it1->start.y=((((x1*y2)-(y1*x2))*(y3-y4))-((y1-y2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4)));
        std::cout<<"linestrip-line intersection x: "<<it->end.x<<" y: "<< it->end.y <<std::endl;
    }
}

void lls_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    std::vector<POINT>::iterator it2;
    it2=it1->control.end();
    //linestrip 1
    double x1=it1->control.back().x;
    double y1=it1->control.back().y;
    double x2=it1->end.x;
    double y2=it1->end.y;
    //line 2
    double x3=it->start.x;
    double y3=it->start.y;
    double x4=it->end.x;
    double y4=it->end.y;

    int determinant=(y2-y4)*(x3-x2)-(x2-x4)*(y3-y2);
    if(determinant==0){
        //skip, lines are collineair
        std::cout<<"linestrip-line is collineair"<<std::endl;
    } else {
        it->end.x=it1->start.x=((((x1*y2)-(y1*x2))*(x3-x4))-((x1-x2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4))); //wikipedia line-line intersection
        it->end.y=it1->start.y=((((x1*y2)-(y1*x2))*(y3-y4))-((y1-y2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4)));
        std::cout<<"linestrip-line intersection x: "<<it->end.x<<" y: "<< it->end.y <<std::endl;
    }
}

void ll_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    //line 1
    double x1=it->start.x;
    double y1=it->start.y;
    double x2=it->end.x;
    double y2=it->end.y;
    //line 2
    double x3=it1->start.x;
    double y3=it1->start.y;
    double x4=it1->end.x;
    double y4=it1->end.y;

    int determinant=(y2-y1)*(x3-x2)-(x2-x1)*(y3-y2);
    if(determinant==0){
        //skip, lines are collineair
    } else {
        it->end.x=it1->start.x=((((x1*y2)-(y1*x2))*(x3-x4))-((x1-x2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4))); //wikipedia line-line intersection
        it->end.y=it1->start.y=((((x1*y2)-(y1*x2))*(y3-y4))-((y1-y2)*(x3*y4-y3*x4))) / (((x1-x2)*(y3-y4))-((y1-y2)*(x3-x4)));
    }
}

void al_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    it->radius=sqrt(pow(it->start.x-it->center.x,2)+pow(it->start.y-it->center.y,2));

    double dx, dy, A, B, C, det, t;
    dx = it1->end.x - it1->start.x;
    dy = it1->end.y - it1->start.y;

    A = dx * dx + dy * dy;
    B = 2 * (dx * (it1->start.x - it->center.x) + dy * (it1->start.y - it->center.y));
    C = (it1->start.x - it->center.x) * (it1->start.x - it->center.x) + (it1->start.y - it->center.y) * (it1->start.y - it->center.y) - it->radius * it->radius;

    det = B * B - 4 * A * C;

    t = (-B + sqrt(det)) / (2 * A);
    double intersection_x1 = (it1->start.x + t * dx);
    double intersection_y1 = (it1->start.y + t * dy);

    t = (-B - sqrt(det)) / (2 * A);
    double intersection_x2 = (it1->start.x + t * dx);
    double intersection_y2 = (it1->start.y + t * dy);

    //find closest solution, at line startpoint..
    double c1 = sqrt(pow(it1->start.x-intersection_x1,2) + pow(it1->start.y-intersection_y1,2));
    double c2 = sqrt(pow(it1->start.x-intersection_x2,2) + pow(it1->start.y-intersection_y2,2));

    if(c1<c2){
        //arc endpoint..
        it->end.x = intersection_x1;
        it->end.y = intersection_y1;
        //line startpoint..
        it1->start.x = intersection_x1;
        it1->start.y = intersection_y1;
    }
    if(c1>c2){
        //arc endtpoint..
        it->end.x = intersection_x2;
        it->end.y = intersection_y2;
        //line startpoint..
        it1->start.x = intersection_x2;
        it1->start.y = intersection_y2;
    }
    //update deg_angle_start, deg_angle_end
    arc_angle_it(it);
}

void la_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    it1->radius=sqrt(pow(it1->start.x-it1->center.x,2)+pow(it1->start.y-it1->center.y,2));

    double dx, dy, A, B, C, det, t;
    dx = it->end.x - it->start.x;
    dy = it->end.y - it->start.y;

    A = dx * dx + dy * dy;
    B = 2 * (dx * (it->start.x - it1->center.x) + dy * (it->start.y - it1->center.y));
    C = (it->start.x - it1->center.x) * (it->start.x - it1->center.x) + (it->start.y - it1->center.y) * (it->start.y - it1->center.y) - it1->radius * it1->radius;

    det = B * B - 4 * A * C;

    t = (-B + sqrt(det)) / (2 * A);
    double intersection_x1 = (it->start.x + t * dx);
    double intersection_y1 = (it->start.y + t * dy);

    t = (-B - sqrt(det)) / (2 * A);
    double intersection_x2 = (it->start.x + t * dx);
    double intersection_y2 = (it->start.y + t * dy);

    //find closest solution, at line startpoint..
    double c1 = sqrt(pow(it->end.x-intersection_x1,2) + pow(it->end.y-intersection_y1,2));
    double c2 = sqrt(pow(it->end.x-intersection_x2,2) + pow(it->end.y-intersection_y2,2));

    if(c1<c2){
        //line endpoint..
        it->end.x = intersection_x1;
        it->end.y = intersection_y1;
        //arc startpoint..
        it1->start.x = intersection_x1;
        it1->start.y = intersection_y1;
    }
    if(c1>c2){
        //line endtpoint..
        it->end.x = intersection_x2;
        it->end.y = intersection_y2;
        //arc startpoint..
        it1->start.x = intersection_x2;
        it1->start.y = intersection_y2;
    }
    arc_angle_it(it1);
}

void aa_intersect(std::vector<OBJECT>::iterator it, std::vector<OBJECT>::iterator it1)
{
    //it->radius=sqrt(pow(it->start.x-it->center.x,2)+pow(it->start.y-it->center.y,2));
    //it1->radius=sqrt(pow(it1->start.x-it1->center.x,2)+pow(it1->start.y-it1->center.y,2));

    // Find the distance between the centers.
    double dx = it->center.x - it1->center.x;
    double dy = it->center.y - it1->center.y;
    double dist = sqrt(dx * dx + dy * dy);

    // See how many solutions there are.
    if (dist > it->radius + it1->radius)
    {
        std::cout<<"No solution"<<std::endl;
    }
    else if (dist < abs(it->radius - it1->radius))
    {
        std::cout<<"No solutions, one circle contains the other"<<std::endl;
    }
    else if ((dist == 0) && (it->radius == it1->radius))
    {
        std::cout<<"No solutions, the circles coincide"<<std::endl;
    }
    else
    {
        // Find a and h.
        double a = (it->radius * it->radius -
                    it1->radius * it1->radius + dist * dist) / (2 * dist);
        double h = sqrt(it->radius * it->radius - a * a);

        // Find P2.
        double cx2 = it->center.x + a * (it1->center.x - it->center.x) / dist;
        double cy2 = it->center.y + a * (it1->center.y - it->center.y) / dist;

        // Get the points P3.
        double px1 = ((cx2 + h * (it1->center.y - it->center.y) / dist));
        double py1 = ((cy2 - h * (it1->center.x - it->center.x) / dist));

        double px2 = (cx2 - h * (it1->center.y - it->center.y) / dist);
        double py2 = ((cy2 + h * (it1->center.x - it->center.x) / dist));

        //calculate the closest by intersection..
        double dist_p1 = sqrt(pow(it->end.x-px1,2)+pow(it->end.y-py1,2));
        double dist_p2 = sqrt(pow(it->end.x-px2,2)+pow(it->end.y-py2,2));

        double dist_p11 = sqrt(pow(it1->start.x-px1,2)+pow(it1->start.y-py1,2));
        double dist_p21 = sqrt(pow(it1->start.x-px2,2)+pow(it1->start.y-py2,2));

        if(dist_p1+dist_p11<dist_p2+dist_p21){
            it->end.x = px1;
            it->end.y = py1;
            it1->start.x = px1;
            it1->start.y = py1;
        } else {
            it->end.x = px2;
            it->end.y = py2;
            it1->start.x = px2;
            it1->start.y = py2;
        }
    }
    arc_angle_it(it);
    arc_angle_it(it1);
}

void arc_angle_it(std::vector<OBJECT>::iterator it){
    double xs=it->start.x;
    double ys=it->start.y;
    double xe=it->end.x;
    double ye=it->end.y;
    double xc=it->center.x;
    double yc=it->center.y;
    double d=it->d;
    if(d>0){
        it->deg_angle_start=180/M_PI*atan2(ye-yc, xe-xc);    //start angle G2
        it->deg_angle_end=180/M_PI*atan2(ys-yc, xs-xc);    //end angle G2
    }
    if(d<0){
        it->deg_angle_start=180/M_PI*atan2(ys-yc, xs-xc);    //start angle G3
        it->deg_angle_end=180/M_PI*atan2(ye-yc, xe-xc);    //end angle G3
    }
    if(d == 0){
        //draw straight line
    }
}

#endif








































